
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include "adlib.h"
#include "pcrc.h"

#include "extname.h"

extern "C" {
   void pcrc_init() ;
   void pcrc_finalize() ;

   int pcrc_new_group_grid(int* rank, int* shape) ;
   void pcrc_delete_group(int* grp) ;

   int pcrc_new_range_distribute(int* g_lb, int* g_ub,
                                  int* distribution, int* grp_grid, int *dim) ;
   int pcrc_new_range_collapse(int* g_lb, int* g_ub) ;
   int pcrc_new_range_align(int* g_lb, int* g_ub,
                             int* offset, int* stride, int* rng_parent) ;
   int pcrc_new_range_copy(int* rng_parent) ;
   int pcrc_new_range_loop(int* g_lb, int* g_ub, int* g_stride,
                            int* offset, int* stride, int* rng_parent) ;
   void pcrc_delete_range(int* rng) ;

   int pcrc_new_array_data(void* array,
                            int* element_type, int* element_size,
                            int* rank, int* grp) ;
   void pcrc_set_array_distribute(int* dad, int* dim, int* g_lb, int* g_ub,
                                   int* distribution,
                                   int* grp_grid, int *grid_dim) ;
   void pcrc_set_array_collapse(int* dad, int* dim,
                                 int* g_lb, int* g_ub) ;
   void pcrc_set_array_align(int* dad, int* dim,
                              int* g_lb, int* g_ub,
                              int* offset, int* stride,
                              int* gl, int* gu,
                              int* rng_parent) ;
   void pcrc_set_array_copy(int* dad, int* dim, int* rng_parent) ;
   void pcrc_set_array_loop(int* dad, int* dim,
                             int* g_lb, int* g_ub, int* g_stride,
                             int* offset, int* stride, int* rng_parent) ;
   
   int pcrc_new_array_section(int* rank, int* dad_parent) ;
   void pcrc_set_array_triplet(int* dad, int* dim,
                                int* lb, int* ub, int* stride,
                                int* dad_parent, int* dim_parent) ;
   void pcrc_set_array_scalar(int* dad,
                               int* i,
                               int* dad_parent, int* dim_parent) ;

   int pcrc_new_array_copy(void* data, int* dad_parent) ;

   void pcrc_delete_array(int* dad) ;

   int pcrc_group(int* dad) ;
   int pcrc_range(int* dad, int* dim) ;
   int pcrc_stride(int* dad, int* dim) ;

   int pcrc_size(int* dad) ;

   void pcrc_reset_array_base(int* dad, void* array) ;

   int pcrc_on(int* grp) ;

   int pcrc_glob(int* rng, int* l_i) ;
   int pcrc_loc(int* rng, int* g_i) ;

   void pcrc_loop_bounds(int* rng, int* l_lb, int* l_ub, int* l_stride) ;

   void pcrc_coef(int * dad, int *dim, int *l, int *u, int *s, int *a, int *b,
                   int *amount, int *m, int *c);

   void pcrc_reset_push_bounds(int* dad, int* dim, int* lb, 
 			       int* lb_parent, int* ub_parent,
                                int* stride_parent) ;
   void pcrc_reset_pop_bounds(int* dad, int* dim,
                               int* lb, int* ub, int* stride) ;

   void pcrc_shift(int* dad, int* dad_source, int* dim, int* amount) ;
   void pcrc_tshift(void* data, int* dad_source, int* dim, int* amount) ;
   void pcrc_remap(int* dad, int* dad_source) ;
   void pcrc_write_halo(int* dad, int* wlo, int* whi) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Global state.                                                       //
//                                                                     //
// Go straight into the Fortran run-time library to pick up the values //
// of `argc' and `argv'.                                               //
//                                                                     //
// There doesn't seem to be a sensible, portable alternative           //
// (this is probably a bug in the MPI definition).                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

#if defined(SOLARIS)

extern "C" {
  extern int __xargc ;
  extern char** __xargv ;
}

static void _AdlibInit() {
  AdlibInit(__xargc, __xargv) ;
}

#elif defined(AIX)

extern "C" {
  extern int p_xargc ;
  extern char** p_xargv ;
}

static void _AdlibInit() {
  AdlibInit(p_xargc, p_xargv) ;
}

#elif defined(NAG)

extern "C" {
  extern int f90_argc ;
  extern char** f90_argv ;
}

static void _AdlibInit() {
  AdlibInit(f90_argc, f90_argv) ;
}

#elif defined(EPC)

extern "C" {
  extern int xargc ;
  extern char** xargv ;
}

static void _AdlibInit() {
  AdlibInit(xargc, xargv) ;
}

#elif defined(MIPS)

extern "C" {
  extern int f77argc ;
  extern char** f77argv ;
}

static void _AdlibInit() {
  AdlibInit(f77argc, f77argv) ;
}

#elif defined(DEC)

extern "C" {
  extern int for__l_argc ;
  extern char** for__a_argv ;
}

static void _AdlibInit() {
  AdlibInit(for__l_argc, for__a_argv) ;
}

#elif defined(F2C)

extern "C" {
  extern int xargc ;
  extern char** xargv ;
}

static void _AdlibInit() {
  AdlibInit(xargc, xargv) ;
}

#else

extern "C" {
  extern int  PCRC_IARGC() ;
  extern void PCRC_GETARG(int*, char*) ;
}

#include <strings.h>

static void getArgs(int &argidx, char *argv[]) {
    argidx = 0;
    char *q;
    int argcnt;

    argcnt = PCRC_IARGC();
    while (argidx <= argcnt)
    {
        q = (char *)malloc(200);
        fflush(stdout);
        PCRC_GETARG(&argidx, q);
        argv[argidx] = q;
        q = (char *)index(argv[argidx],' ');
        *q = '\0';
        ++argidx;
    }
}

static void _AdlibInit() {
  int argidx;
  char * argv[200];
  char **argvp = argv;

  getArgs(argidx, argv);
  AdlibInit(argidx, argvp) ;
}

#endif

static Group* rootPtr ;

static int ad_FromPointer(const void* ptr) ;
static Procs* getProcs(int rank, int shape []) ;

void pcrc_init() {
  _AdlibInit() ;

  ad_FromPointer(&apg) ;     // should return 1

  ad_FromPointer(new Group(*getProcs(0, (int*) 0))) ;  // should return 2
}

void pcrc_finalize() {
  AdlibFinalize() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Routines to convert to and from pointers.                          //
//                                                                     //
//  This code is lifted, with minimal modification, from               //
//  `mpich/src/util/ptrcvt.c'.  It is a dirty fix for the fact that    //
//  pointers and int's are different lengths on the DEC Alpha.         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Ad_PtrToIdx {
  int idx;
  const void *ptr;
  Ad_PtrToIdx *next;
} ;

#define MAX_PTRS 10000

static Ad_PtrToIdx PtrArray[MAX_PTRS];
static Ad_PtrToIdx *avail=0;
static int      DoInit = 1;

static void ad_InitPointer() {
  int  i;
  for (i=0; i<MAX_PTRS-1; i++) {
    PtrArray[i].next = PtrArray + i + 1;
    PtrArray[i].idx  = i;
  }
  PtrArray[MAX_PTRS-1].next = 0;
  /* Don't start with the first one, whose index is 0. That could
     break some code. */
  avail   = PtrArray + 1;
}

static const void *ad_ToPointer(int idx) {
  if (DoInit) {
    DoInit = 0;
    ad_InitPointer();
  }
  if (idx < 0 || idx >= MAX_PTRS) {
    fprintf(stderr, "** RUNTIME ERROR **  "
            "Could not convert index %d into a pointer\n", idx );
    exit(1);
  }
  if (idx == 0) return (void *)0;
  return PtrArray[idx].ptr;
}

static int ad_FromPointer(const void* ptr) {
  int      idx;
  Ad_PtrToIdx *newP;
  if(DoInit) {
    DoInit = 0;
    ad_InitPointer();
  }
  if(!ptr) return 0;
  if(avail) {
    newP	      = avail;
    avail	      = avail->next;
    newP->next	      = 0;
    idx		      = newP->idx;
    PtrArray[idx].ptr = ptr;
    return idx;
  }
  // This isn't the right thing to do, but it isn't too bad
  fprintf(stderr, "** RUNTIME ERROR **  Pointer conversions exhausted\n" );
  exit(1);
  // Some systems may complain here about the lack of a return.
  return 0;
}

static void ad_RmPointer(int idx) {
  int myrank;
  if (DoInit) {
    DoInit = 0;
    ad_InitPointer();
    }
  if(idx < 0 || idx >= MAX_PTRS) {
    fprintf(stderr, "** RUNTIME ERROR **  "
            "Could not convert index %d into a pointer\n", idx );
    exit(1);
    }
  if (idx == 0) return;
  if (PtrArray[idx].next) {
    /* In-use pointers NEVER have next set */
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
    fprintf(stderr, "** RUNTIME ERROR **  "
	    "[%d] Error in recovering Fortran pointer; already freed\n", 
	    myrank );
    exit(1);
  }
  PtrArray[idx].next = avail;
  PtrArray[idx].ptr  = 0;
  avail              = PtrArray + idx;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Processor grid.                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// `processArrays' table is used to ensure that if several process
// grids are defined with the same shape they refer to the same
// `Procs' structure.  It shouldn't be confused with the handle-to-pointer
// translation table `PtrArray'.

// Construction and destruction.

struct Ad_ProcessArrayRecord {
  int rank ;
  int shape [7] ;
  Procs* procs ;
} ;

#define MAX_PROCESS_ARRAYS 100

static Ad_ProcessArrayRecord processArrays [MAX_PROCESS_ARRAYS] ;
static int numProcessArrays = 0 ;

static Procs* getProcs(int rank, int shape []) {

  // Scan `processArrays' table for pre-existing process array of the
  // same shape.  Return that if it exists.

  for(int slot = 0 ; slot < numProcessArrays ; slot++) {
    Ad_ProcessArrayRecord& processArray = processArrays [slot] ;
    if(rank == processArray.rank) {
      int i = 0 ;
      while (i < rank && shape [i] == processArray.shape [i]) i++ ;
      if(i == rank) return processArray.procs;
    }
  }

  // Otherwise create a new `Procs'

  if(numProcessArrays == MAX_PROCESS_ARRAYS) {
    fprintf(stderr, "** RUNTIME ERROR **  Process array records exhausted\n") ;
    exit(1) ;
    return 0 ;
  }
  else {
    int i ;

    // Check size of processor array is number of available processors.

    int size = 1 ;
    for (i = 0 ; i < rank ; i++)
      size *= shape [i] ;

    int asize = apg.size() ;

    if(size != asize && size != 1) {
      fprintf(stderr, "** RUNTIME ERROR **  "
              "Processor array requires %d processes---only %d available\n",
              size, asize) ;
      exit(1) ;
    }

    // Create new process array, and add it to `processArrays' table.

    Ad_ProcessArrayRecord& processArray = processArrays [numProcessArrays] ;

    processArray.rank = rank ;
    for (i = 0 ; i < rank ; i++)
      processArray.shape [i] = shape [i] ;
    processArray.procs = new Procs(rank, shape) ;

    numProcessArrays++ ;

    return processArray.procs ;
  }
}

int pcrc_new_group_grid(int* rank, int* shape) {
  return ad_FromPointer(new Group(*getProcs(*rank, shape))) ;
}

void pcrc_delete_group(int* grp) {
  delete (Group*) ad_ToPointer(*grp) ;

  ad_RmPointer(*grp) ;
}

int pcrc_on(int* grp) {
  Group* p = (Group*) ad_ToPointer(*grp) ;

  return p->member() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Initialization of ranges                                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void setDistribute(Range& rng, const int _g_lb, const int _g_ub,
                   const int distribution, Group* grpPtr, const int dim) {

  Dimension d = grpPtr->prc()->dim(dim - 1) ;
  const int extent = _g_ub - _g_lb + 1 ;
  switch(distribution) {
    case 1 :
      rng = BlockRange(extent, d) ;
      break ;
    case 2 :
      rng = CyclicRange(extent, d) ;
      break ;
  }

  rng.g_lb()     = _g_lb ;
  rng.g_stride() = 1 ;
}

void setCollapse(Range& rng, const int _g_lb, const int _g_ub) {

  const int extent = _g_ub - _g_lb + 1 ;
  rng = Range(extent) ;

  rng.g_lb()     = _g_lb ;
  rng.g_stride() = 1 ;
}

void setAlign(Range& rng, const int _g_lb, const int _g_ub,
              const int a, const int b, const int gl, const int gu,
              Range parent) {

  const int extent = _g_ub - _g_lb + 1 ;
  const int base   = (a * _g_lb + b - parent.g_lb()) / parent.g_stride() ;
  const int stride = a / parent.g_stride() ;

  if(gl == 0 && gu == 0)
    rng = parent.subrng(extent, base, stride) ;
  else {
    // Create range with SHADOW layout.

    const int parBas = parent.bas() ;
    const int parStr = parent.str() ;

    BlockRange x(parent.tem().size(), parent.dim(), gl, gu) ;
    rng = x.subrng(extent, parBas + parStr * base, parStr * stride) ;
  }

  rng.g_lb()     = _g_lb ;
  rng.g_stride() = 1 ;
}

void setLoop(Range& rng, const int _g_lb, const int _g_ub, const int _g_stride,
             const int a, const int b, Range parent) {

  const int extent = (_g_ub - _g_lb) / _g_stride + 1 ;
  const int base   = (a * _g_lb + b - parent.g_lb()) / parent.g_stride() ;
  const int stride = a * _g_stride / parent.g_stride() ;
  rng = parent.subrng(extent, base, stride) ;

  rng.g_lb()     = _g_lb ;
  rng.g_stride() = _g_stride ;
}

void setTriplet(Range& rng, const int lb, const int ub, const int stride,
                Range parent) {

  int extent  = (ub - lb) / stride + 1 ;
  int base    = (lb - parent.g_lb()) / parent.g_stride() ;
  int stride_ = stride / parent.g_stride() ;
  rng = parent.subrng(extent, base, stride_) ;

  rng.g_lb()     = lb ;
  rng.g_stride() = stride ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Range creation.                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int pcrc_new_range_distribute(int* g_lb, int* g_ub,
                              int* distribution, int* grp_grid, int* dim) {

  Range* rng = new Range ;
  Group* grpPtr = (Group*) ad_ToPointer(*grp_grid) ;
  setDistribute(*rng, *g_lb, *g_ub, *distribution, grpPtr, *dim) ;

  return ad_FromPointer(rng) ;
}

int pcrc_new_range_collapse(int* g_lb, int* g_ub) {

  Range* rng = new Range ;
  setCollapse(*rng, *g_lb, *g_ub) ;

  return ad_FromPointer(rng) ;
}

int pcrc_new_range_align(int* g_lb, int* g_ub,
                          int* a, int* b, int* rng_parent) {
  Range* rng = new Range ;
  setAlign(*rng, *g_lb, *g_ub, *a, *b, 0, 0,
           *((Range*) ad_ToPointer(*rng_parent))) ;

  return ad_FromPointer(rng) ;
}

int pcrc_new_range_loop(int* g_lb, int* g_ub, int* g_stride,
                         int* a, int* b, int* rng_parent) {
  Range* rng = new Range ;
  setLoop(*rng, *g_lb, *g_ub, *g_stride, *a, *b,
          *((Range*) ad_ToPointer(*rng_parent))) ;

  return ad_FromPointer(rng) ;
}

int pcrc_new_range_copy(int* rng_parent) {
  Range* rng = new Range ;

  *rng = *((Range*) ad_ToPointer(*rng_parent)) ;

  return ad_FromPointer(rng) ;
}

void pcrc_delete_range(int* handle) {
  Range* rng = (Range*) ad_ToPointer(*handle) ;

  ad_RmPointer(*handle) ;
  delete rng ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class pcrcfDAD.                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

pcrcfDAD :: pcrcfDAD(void* _data,
                     const int _elementType, const int _elementLen,
                     const int _rank, const Group& _group) :
                DAD(_rank, _group),
                elementLen(_elementLen), elementType(_elementType),
                size(1), data(_data) {
  DAD :: ranges  = rngs ;
  DAD :: strides = strs ;
  grpHandle   = ad_FromPointer(&(group)) ;
  for(int r = 0 ; r < rank ; r++)
    rngHandles [r] = ad_FromPointer(&(ranges [r])) ;
}


pcrcfDAD :: pcrcfDAD(const int _rank, pcrcfDAD& parent)  :
                DAD(_rank, parent.grp()),
                elementLen(parent.elementLen), elementType(parent.elementType),
                size(parent.size), data(parent.data) {
  DAD :: ranges  = rngs ;
  DAD :: strides = strs ;
  grpHandle   = ad_FromPointer(&(group)) ;
  for(int r = 0 ; r < rank ; r++)
    rngHandles [r] = ad_FromPointer(&(ranges [r])) ;
}

pcrcfDAD :: pcrcfDAD(void* _data, pcrcfDAD& parent)  :
                DAD(parent),
                elementLen(parent.elementLen), elementType(parent.elementType),
                size(parent.size), data(_data) {
  DAD :: ranges  = rngs ;
  DAD :: strides = strs ;
  grpHandle   = ad_FromPointer(&(group)) ;
  for(int r = 0 ; r < rank ; r++)
    rngHandles [r] = ad_FromPointer(&(ranges [r])) ;
}

pcrcfDAD :: ~pcrcfDAD() {
  ad_RmPointer(grpHandle) ;
  for(int r = 0 ; r < rank ; r++)
    ad_RmPointer(rngHandles [r]) ;
}

void pcrcfDAD :: addRange(const int r, Range x) {
  ranges [r] = x ;

  strides [r] = size ;
  size *= x.volume() ;
}

void pcrcfDAD :: addTriplet(const int r,
                            const int lb, const int ub, const int stride,
                            pcrcfDAD& parent, const int s) {
  Range x ;
  setTriplet(x, lb, ub, stride, parent.rng(s)) ;

  ranges  [r] = x ;
  strides [r] = parent.jmp(s) ;
}

void pcrcfDAD :: addScalar(const int u, pcrcfDAD& parent, const int s) {
  Range y = parent.rng(s) ;

  const int glb = (u - y.g_lb()) / y.g_stride() ;
  //int sub ;
  //const int crd = y.ker_glb(glb, &sub) ;
  Location i(y(glb)) ;

  //return l_i ;

  //(char*&) data += elementLen * parent.str(s).offset(i) ;
  (char*&) data += elementLen * parent.jmp(s) * y.disp(i.sub) ;
  group.restrict(y.dim(), i.crd) ;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Array creation.                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int pcrc_new_array_data(void* data,
                        int* element_type, int* element_size,
                        int* rank, int* grp) {
  pcrcfDAD* ptr = new pcrcfDAD(data, *element_type, *element_size,
                     *rank, *((Group*) ad_ToPointer(*grp))) ;

  return ad_FromPointer(ptr) ;
}

void pcrc_set_array_distribute(int* dad, int* dim, int* g_lb, int* g_ub,
                               int* distribution,
                               int* grp_grid, int *grid_dim) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  Range x ;
  setDistribute(x, *g_lb, *g_ub, *distribution,
                (Group*) ad_ToPointer(*grp_grid), *grid_dim) ;

  ptr->addRange(*dim - 1, x) ;
}

void pcrc_set_array_collapse(int* dad, int* dim, int* g_lb, int* g_ub) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  Range x ;
  setCollapse(x, *g_lb, *g_ub) ;

  ptr->addRange(*dim - 1, x) ;
}

void pcrc_set_array_align(int* dad, int* dim,
                          int* g_lb, int* g_ub,
                          int* a, int* b,
                          int* gl, int* gu,
                          int* rng_parent) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  Range x ;
  setAlign(x, *g_lb, *g_ub, *a, *b, *gl, *gu,
           *((Range*) ad_ToPointer(*rng_parent))) ;

  ptr->addRange(*dim - 1, x) ;
}

void pcrc_set_array_copy(int* dad, int* dim, int* rng_parent) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  ptr->addRange(*dim - 1, *((Range*) ad_ToPointer(*rng_parent))) ;
}


void pcrc_delete_array(int* dad) {
  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;

  ad_RmPointer(*dad) ;

  delete ptr ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Section creation.                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int pcrc_new_array_section(int* rank, int* dad_parent) {
  pcrcfDAD* ptr = new pcrcfDAD(*rank, *((pcrcfDAD*) ad_ToPointer(*dad_parent))) ;

  return ad_FromPointer(ptr) ;
}


void pcrc_set_array_triplet(int* dad, int* dim,
			    int *lb, int *ub, int *stride,
			    int* dad_parent, int* dim_parent) {

  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;

  ptr->addTriplet(*dim - 1, *lb, *ub, *stride,
                  *((pcrcfDAD*) ad_ToPointer(*dad_parent)), *dim_parent - 1) ;
}

void pcrc_set_array_scalar(int* dad,
		           int *u,
		           int* dad_parent, int* dim_parent) {

  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;

  ptr->addScalar(*u, *((pcrcfDAD*) ad_ToPointer(*dad_parent)), *dim_parent - 1) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Array copying.                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int pcrc_new_array_copy(void* data, int* dad_parent) {
  pcrcfDAD* ptr = new pcrcfDAD(data, *((pcrcfDAD*) ad_ToPointer(*dad_parent))) ;

  return ad_FromPointer(ptr) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Misc operations on arrays.                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int pcrc_group(int* dad) {
  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;

  return ptr->grpHandle ;
}

int pcrc_range(int* dad, int* dim) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  return ptr->rngHandles [*dim - 1] ;
}

int pcrc_stride(int* dad, int* dim) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  //return ptr->str(*dim - 1).len() ;
  return ptr->jmp(*dim - 1) ;
}

int pcrc_size(int* dad) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  return ptr->size ;
}


void pcrc_reset_array_base(int* dad, void* data) {
  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;

  ptr->data = data ;
}


void pcrc_reset_push_bounds(int* dad, int* dim, int* lb,
                            int* lb_parent, int* ub_parent,
                            int* stride_parent ) {

  // (Last argument now ignored.)

  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;
  Range x  = ptr->rng(*dim - 1) ;

  *lb_parent     = x.g_lb() ;
  *stride_parent = x.g_stride() ;

  x.g_lb()     = *lb ;
  x.g_stride() = 1 ;
}

void pcrc_reset_pop_bounds(int* dad, int* dim,
                           int* lb_parent, int* ub_parent, int* stride_parent) {

  // (Last argument now ignored.)

  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;
  Range x  = ptr->rng(*dim - 1) ;

  x.g_lb()     = *lb_parent ;
  x.g_stride() = *stride_parent ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Address translation.                                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int pcrc_glob(int* rng, int* l_i) {
  Range x = *((Range*) ad_ToPointer(*rng)) ;

  return x.g_lb() + x.glb(*l_i, x.dim().crd()) * x.g_stride() ;
}

int pcrc_loc(int* rng, int* g_i) {
  Range x = *((Range*) ad_ToPointer(*rng)) ;
  //int l_i ;

  const int glb = (*g_i - x.g_lb()) / x.g_stride() ;
  //x.ker_glb(glb, &l_i) ;
  //return l_i ;

  Location i(x(glb)) ;
  return i.sub ;
}

void pcrc_loop_bounds(int* rng, int* l_lb, int* l_ub, int* l_stride) {
  Range x = *((Range*) ad_ToPointer(*rng)) ;

  // Parameters of local block of `x'.

  int count ;

  x.block(x.dim().crd(), l_lb, &count) ;
  x.block_step(l_stride) ;
  *l_ub = *l_lb + *l_stride * (count - 1) ;
}

void pcrc_coef(int *dad, int *dim, int *l, int *k, int *s, int *a, int *b,
               int *amount, int *m, int *c) {

  const int r = *dim - 1;
  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;
  Range x = ptr->rng(r) ;

  const int t_stride = x.str() / x.g_stride() ;

  // Assume an imaginary, shifted version of the data, moved `amount'
  // places to the left in the template.

  // The subrange `y' defines the section of the imaginary array
  // selected by the loop bounds.

  const int extent = (*k - *l) / *s + 1 ;
  const int base   = x.bas() - *amount + t_stride * (*a * *l + *b - x.g_lb()) ;
  const int stride = t_stride * *a * *s ;

  Range y = x.tem().subrng(extent, base, stride) ;

  // Parameters of local block of `y'.

  int count, sub, sub_stp ;
  y.block(y.dim().crd(), &sub, &count) ;
  y.block_step(&sub_stp) ;

  // Convert to local subscript.  Actual data comes from `amount'
  // places to the right of where it would be in the imaginary array
  // (ie, go into ghost regions of actual data).

  //Stride u = ptr->str(r) ;

  //*m = u.disp_step(sub_stp) ;
  //*c = u.disp(sub) + u.len() * *amount ;

  const int itemLen = ptr->jmp(r) ;

  *m = itemLen * x.step(sub_stp) ;
  *c = itemLen * (x.disp(sub) + *amount) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Communication.                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void pcrc_shift(int* dad, int* dad_source, int* dim, int* amount) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_source) ;

  Shift schedule(ptr, srcPtr, ptr->elementLen, *amount, *dim - 1, EDGE) ;
  schedule.execute(ptr->data, srcPtr->data) ;
}


void pcrc_tshift(void* data, int* dad_source, int* dim, int* amount) {

  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_source) ;
  int d = *dim - 1 ;

  // Create `DAD' record containing a primitive range in
  // place of the actual source array range.

  const int rank = srcPtr->rnk() ;
  Range  rngs [8] ;
  int strs [8] ;
  for(int r = 0 ; r < rank ; r++) {
    if(r == d)
      rngs [r] = srcPtr->rng(r).tem() ;
    else
      rngs [r] = srcPtr->rng(r) ;
    strs [r] = srcPtr->jmp(r) ;
  }
  DAD tmp(rank, srcPtr->grp(), rngs, strs) ;

  // Do the shift

  Shift schedule(&tmp, &tmp, srcPtr->elementLen, *amount, d, EDGE) ;
  schedule.execute(data, srcPtr->data) ;
}


void pcrc_remap(int* dad, int* dad_source) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_source) ;

  Remap schedule(ptr, srcPtr, ptr->elementLen) ;
  schedule.execute(ptr->data, srcPtr->data) ;
}


void pcrc_write_halo(int* dad, int* wlo, int* whi) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  const int rank = ptr->rnk() ;
  int* w = new int [2 * rank] ;

  int* wloAdj = w ; 
  int* whiAdj = w + rank ; 

  for(int r = 0 ; r < rank ; r++) {
    const int str = ptr->rng(r).str() ;
    wloAdj [r] = (wlo [r] + str - 1) / str ;
    whiAdj [r] = (whi [r] + str - 1) / str ;
  }

  WriteHalo schedule(ptr, ptr->elementLen, wloAdj, whiAdj) ;
  schedule.execute(ptr->data) ;

  delete [] w ;
}


// -------------------------------------------------------------
// Things to do
//
//   Delete `tshift_', and remove the `amount' argument from `pcrc_coef_'.
//
//   Negotiate to lose interim `g_lb', `g_stride' members on `Range',
//       thus lose `pcrc_reset_push_bounds_', `pcrc_reset_pop_bounds_'.
//
//   Layer this as a Fortran-callable wrapper around a C-callable interface?
//
//   Finally (?) sort out the argc, argv business.
//
//   Move `extern' function prototypes into `pcrc.h'.

